"""
Abstract base class for `exact` routers.

(c) Copyright Bprotocol foundation 2022.
Licensed under MIT
"""
from abc import ABC

from .base_router import *


@dataclass
class ExactBase(BaseRouter, ABC):
    """
    Abstract base class for both ExactRouterSBY and ExactRouterX0Y0N routers.
    """

    def dxdy(self, subject: int, dy: DecFloatInt) -> DecFloatInt:
        """
        Calculates delta-x (dx) contribution for y
        """
        x0 = self.orders[subject].x0
        y0 = self.orders[subject].y0
        n = self.orders[subject].n
        y = self.orders[subject].y
        self.logger.debug(
            f"y: {self.frmt(y)}, dy: {self.frmt(dy)}, y0: {self.frmt(y0)}, x0: {self.frmt(x0)}, n: {self.frmt(n)} \n"
        )
        return -x0 * y0 * dy / ((n * y + y0 * (1 - n)) * (n * (y + dy) + y0 * (1 - n)))

    @staticmethod
    def check_partial_swaps(partial_swaps):
        """
        Check whether trade routing is complete.
        """
        if partial_swaps:
            return all(swap <= 0 for swap in partial_swaps.values())
        else:
            return False

    def match_by_src(
            self,
            x: DecFloatInt,
            is_by_target: bool = False,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = 100,
    ) -> List[Action]:
        """
        Alias for match method in the case of a source amount.
        """
        return self.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
        )

    def match_by_target(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = 100,
    ) -> List[Action]:
        """
        Alias for match method in the case of a target amount.
        """
        return self.match(
            x=x,
            is_by_target=is_by_target,
            check_sufficient_liquidity=check_sufficient_liquidity,
        )

    def match(
            self,
            x: DecFloatInt,
            is_by_target: bool = True,
            completed_trade: bool = False,
            trade: Callable = None,
            cmp: Callable = None,
            check_sufficient_liquidity: bool = True,
            threshold_orders: int = 100,
            use_positions_matchlevel: List[int] = [],
    ) -> List[Action]:
        """
        Main logic for the exact methods (both SBY and Y0X0N variable implementations).
        This method determines the ideal distribution of the input, and calculates the corresponding output.
        """

        if is_by_target:
            if check_sufficient_liquidity:
                self.sufficient_liquidity_exists(x, use_positions_matchlevel)
        partial_swaps = final_swaps = {}
        dropped_indices = []
        while not completed_trade:
            if len(use_positions_matchlevel) == 0:
                dropped_indices = list(final_swaps.keys())
            else:
                dropped_indices = [i for i in self.indexes if i not in use_positions_matchlevel] + list(final_swaps.keys())

            if is_by_target:
                delta = sum([i for i in final_swaps.values()])
            else:
                delta = sum([self.dxdy(k, v) for k, v in final_swaps.items()])

            amt = x - delta

            self.logger.info(
                f"amt: {self.frmt(amt)}, delta: {self.frmt(delta)}, x: {x}, dropped_indices: {dropped_indices}, final_swaps: {[(k, self.frmt(v)) for k, v in final_swaps.items()]} \n"
            )

            self.logger.info(
                f"amt: {self.frmt(amt)}, delta: {self.frmt(delta)}, x: {x} \n"
            )

            partial_swaps = {}
            while not self.check_partial_swaps(partial_swaps):
                dropped_indices += [k for k, v in partial_swaps.items() if v > 0] + [
                    i for i in self.indexes if self.orders[i].y == 0
                ]
                position_subset = [i for i in self.indexes if i not in dropped_indices]
                if len(position_subset) == 0:
                    break

                partial_swaps.clear()
                for i in position_subset:
                    if is_by_target:
                        partial_swap_amount = self.amt_by_target(
                            i, amt, position_subset
                        )
                    else:
                        partial_swap_amount = self.amt_by_src(i, amt, position_subset)

                    self.logger.debug(
                        f"partial_swap_amount: {i}, {self.frmt(partial_swap_amount)} \n"
                    )

                    partial_swaps[i] = partial_swap_amount

            completed_trade = True
            for index, trade_target in partial_swaps.items():
                position_liquidity = self.orders[index].y
                test_liquidity = position_liquidity + trade_target < 0

                self.logger.debug(
                    f"index: {index}, test_liquidity: {test_liquidity}, position_liquidity: {self.frmt(position_liquidity)}, trade_target: {self.frmt(trade_target)}\n"
                )

                if test_liquidity:
                    final_swaps[index] = -position_liquidity
                    completed_trade = False

            self.logger.debug(
                f"final_swaps: {[(k, self.frmt(v)) for k, v in final_swaps.items()]} \n"
            )
        
        # Format Results
        final_swaps = {**final_swaps, **partial_swaps}
        outputs = {k: v for k, v in final_swaps.items() if v != 0}
        inputs = {k: self.dxdy(k, v) for k, v in outputs.items()}
        ttl_values = {}
        ttl_inputs = {}
        indices = [i for i in inputs.keys()]
        indexes = [i for i in range(len(indices))]
        for i in indexes:
            ttl = outputs[indices[i]]
            inpts = inputs[indices[i]]
            if i > 0:
                ttl_values[i] = ttl_values[i - 1] + ttl
                ttl_inputs[i] = ttl_inputs[i - 1] + inpts
            else:
                ttl_values[i] = ttl
                ttl_inputs[i] = inpts
        sum_ttl_inputs = sum([v for k,v in ttl_inputs.items()])

        # assert that in and out match; they should, but for numerical reasons there may be small
        # differences, which we tolerate with a warning; however, there are some error conditions 
        # that lead to bigger changes and we want to react appropriately here
        goal_value = float(x)

        # if we are trading by target, we compare against the sum of the outputs, else sum the inputs
        actual_value = float(
            sum([v for v in outputs.values()])
        ) if is_by_target else float(
            sum([v for v in inputs.values()])
        )
        perc_error = abs(goal_value / actual_value - 1)
        msg = f"matching finished: err={perc_error}, " \
              f"goal={goal_value}, " \
              f"actual={actual_value}, " \
              f"is_by_target={is_by_target}"
        self.logger.debug(msg)
        #print("[match]", msg)

        if not is_by_target:
            try:
                assert (
                        float(actual_value) - float(goal_value) >= 0
                ), "Insufficient liquidity across all user positions to support this trade."
            except AssertionError as e:
                raise e

        if perc_error > 0.0001:
            err_message = f"[match] imprecise matching: err={perc_error} (goal={goal_value}, actual={actual_value}, is_by_target={is_by_target}) "
            print(err_message)
            if perc_error < 0.01:
                self.logger.info(err_message)       # message logged as info <1%...
            elif perc_error < 0.05:
                self.logger.warning(err_message)    # ...as warning 1-5%...
            else:
                self.logger.error(err_message)      # ...and error/raise >5%
                raise RuntimeError(err_message)
            assert perc_error < 0.10, err_message

        # Package Results
        match_method = "by_src" if not is_by_target else "by_target"
        actions = [
            Action(
                match_method=match_method,
                index=indices[i],
                x=x,
                input=self.frmt(inputs[indices[i]]),
                output=self.frmt(outputs[indices[i]]),
                total_output=self.frmt(ttl_values[i]),
                total_input=self.frmt(ttl_inputs[i]),
                threshold_orders=threshold_orders,
            )
            for i in indexes
        ]
        return actions



"""
represents a Carbon order book position

(c) Copyright Bprotocol foundation 2022.
Licensed under MIT
"""
import pprint
from dataclasses import dataclass
from decimal import Decimal
from decimal import getcontext
from math import sqrt
from typing import NewType

from .pair import CarbonPair


getcontext().prec = 128

pp = pprint.PrettyPrinter(indent=4)

DecFloatInt = NewType("value", int or Decimal or float)


@dataclass
class Order:
    """
    Curve class to represent a curve on the market maker's order book
    """

    # Production variables
    B: DecFloatInt = None  # = sqrt(price_low)
    S: DecFloatInt = None  # = sqrt(price_high) - sqrt(price_low)
    y_int: DecFloatInt = None  # is the initial intercept of the curve

    # Original variables
    p_high: DecFloatInt = None  # = high price, alias "p_a" in original code/docs
    p_low: DecFloatInt = None  # = low price, alias "p_a" in original code/docs
    _y: DecFloatInt = None  # is the current balance of the curve

    # Alternate variables
    D: DecFloatInt = None  # is the current intercept of the curve
    _C: DecFloatInt = None  # is the current balance of the curve

    pair: CarbonPair = None
    tkn: str = None
    pair_name: str = None
    adj_n_mode: bool = False
    auto_convert_variables: bool = True
    id: int = 0
    linked_to_id: int = 0
    _reverseq: bool = False
    verbose: bool = False

    @property
    def y(self) -> Decimal:
        """
        The current balance of the curve... Equal to y_int upon genesis.
        """
        return self.y_int if self._y is None else self._y

    @y.setter
    def y(self, value: Decimal) -> None:
        """
        Setter for the current balance of the curve...
        """
        self._y = value

    @property
    def C(self) -> Decimal:
        """
        The current balance of the curve... Equal to D upon genesis.
        """
        return self.D if self._C == 0 else self._C

    @C.setter
    def C(self, value: Decimal) -> None:
        """
        Setter for the current balance of the curve...
        """
        self._C = value

    @property
    def p(self) -> Decimal:
        p_high: Decimal = self.p_high
        p_low: Decimal = self.p_low
        return (p_high * p_low).sqrt()

    @property
    def x0(self) -> Decimal:
        n: Decimal = self.n
        p: Decimal = self.p
        y_int: Decimal = max(self.y_int, Decimal("1"))
        return (y_int * (1 - n)) / (p * (2 - n))

    @property
    def k(self) -> Decimal:
        x0 = self.x0
        y0 = self.y0
        return x0 * y0

    @property
    def a(self):
        return self.k * (self.n - 1) - self.n * self.x0 * self.y

    @property
    def x(self) -> Decimal:
        x0: Decimal = self.x0
        n: Decimal = self.n
        y: Decimal = self.y
        y0: Decimal = self.y0
        return x0 * (n * y - y + y0 * (2 - n)) / (n * y + y0 * (1 - n))

    @property
    def y0(self) -> Decimal:
        y_int: Decimal = max(self.y_int, Decimal("1"))
        n: Decimal = self.n
        return (y_int * (1 - n)) / (2 - n)

    @property
    def n(self) -> Decimal:
        p_high: Decimal = self.p_high
        p_low: Decimal = self.p_low
        return (
            1 - (p_low / p_high) ** Decimal(".25")
            if not self.adj_n_mode
            else Decimal("0.000000000000000001")
        )

    @property
    def reverseq(self):
        return self._reverseq

    @reverseq.setter
    def reverseq(self, value):
        self._reverseq = value

    # Use the dataclass post-init method to initialize all variable sets
    def __post_init__(self):

        if self.p_high is not None and self.p_low is not None:
            self.p_high = Decimal(self.p_high)
            self.p_low = Decimal(self.p_low)
            self.y_int = Decimal(self.y_int)

        if self.auto_convert_variables:

            # Init original variables if alternate variables are set
            if self.B is not None and self.S is not None:
                self.B = Decimal(self.B)
                self.S = Decimal(self.S)
                self.pb_raw = self.B * self.B
                self.pa_raw = (self.S + self.B) ** 2
                self.pa = 1 / self.pa_raw if self.reverseq else self.pa_raw
                self.pb = 1 / self.pb_raw if self.reverseq else self.pb_raw
                if self.pa < self.pb:
                    self.p_low = self.pa
                    self.p_high = self.pb
                else:
                    self.p_low = self.pb
                    self.p_high = self.pa

            # Init alternate variables if original variables are set
            elif self.p_low is not None and self.p_high is not None:

                if self.p_high < self.p_low:
                    paa = self.p_high
                    self.p_high = self.p_low
                    self.p_low = paa

                self.B = sqrt(self.p_low)
                self.S = sqrt(self.p_high) - sqrt(self.p_low)

            if self.y_int is not None:
                self.D = self.C = self._y = self.y_int

            elif self.D is not None:
                self.y_int = self.C = self._y = self.D

            if self.p_high == self.p_low:
                self.adj_n_mode = True



"""
represents a single, unidirectional carbon order and provides convenience methods for UI access

(c) Copyright Bprotocol foundation 2022. 
Licensed under MIT
"""
__version__ = "1.1"
__date__ = "24/Nov/2022"

try:
    from .pair import CarbonPair
except:
    from pair import CarbonPair

from dataclasses import dataclass
from math import sqrt

@dataclass
class CarbonOrderUI:
    """
    current state of a single Carbon order, for UI purposes
    
    properties that are parameters of the constructor
    :pair:    the corresponding token pair (specifically, its CarbonPair record)
    :tkn:     the token that this position is selling
    :B:       the B-parameter; B = sqrt pb_raw
    :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw
    :yint:    the y-intercept of the curve (also its current maximum capacity)
    :y:       the current y-coordinate on the curve (also current token holdings)
    other properties
    :pa_raw:   the pa parameter in native quotation (dy/dx)
    :pb_raw:   ditto pb
    :pa:       the pa paramter in the quotation appropriate for the pair
    :pb:       ditto pb
    :pmin:     the min of pa, pb (in the quotation appropriate for the pair)
    :pmax:     the max of pa, pb (in the quotation appropriate for the pair)
    
    NOTES
    - natively, prices are quoted in the convention dy/dx, where tkn is the quote asset;
    tkn is also the asset being sold, so the numeraire is always the asset being sold
    - pa, pb are read at the intercepts "left to right", so pa=py is is the y intercept price,
    and pb=px is the x intercept price 
    - the properties pa_raw and pb_raw correspond to the native pa, pb; the prperties pa, pb
    are quoted in the correct currency conventions
    - the properties py = pa and px = pb are aliases
    """
    __VERSION__ = "1.0"
    
    pair: CarbonPair
    tkn: str
    B: float
    S: float
    yint: float
    y: float
    def __post_init__(self):
        self.tkn = self.tkn.upper()
        if not self.pair.has_token(self.tkn):
            raise RuntimeError("token not part of pair", self.tkn, self.pair)
        self.pb_raw = self.B * self.B
        self.pa_raw = (self.S + self.B)**2
        self.reverseq = True if self.pair.has_basetoken(self.tkn) else False
        self.pa = 1./self.pa_raw if self.reverseq else self.pa_raw
        self.pb = 1./self.pb_raw if self.reverseq else self.pb_raw
        if self.pa < self.pb:
            self.pmin = self.pa
            self.pmax = self.pb
        else:
            self.pmin = self.pb
            self.pmax = self.pa          
    
    @classmethod
    def from_BSy(cls, pair, tkn, B, S, yint, y):
        """
        alternative* constructor, taking B,S,y

        :pair:    the corresponding token pair (specifically, its CarbonPair record)
        :tkn:     the token that this order is selling
        :B:       the B-parameter; B = sqrt pb_raw
        :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw
        :yint:    the y-intercept of the curve (also its current maximum capacity)
        :y:       the current y-coordinate on the curve (also current token holdings)

        *technically this is the same as the constructor, but use of `from_BSy` is recommended
        over use of the native class constructor in case of future implementation changes
        """
        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S= S, 
            yint=yint, 
            y=y
        )

    @classmethod
    def from_prices(cls, pair, tkn, pa, pb, yint, y):
        """
        alternative constructor, taking prices pa, pb
        
        :pair:    the corresponding token pair (specifically, its CarbonPair record)
        :tkn:     the token that this order is selling
        :pa:      the price at the y intercept, in quotation corresponding to the pair*
        :pb:      the price at the x intercept, in quotation corresponding to the pair*
        :yint:    the y-intercept of the curve (also its current maximum capacity)
        :y:       the current y-coordinate on the curve (also current token holdings)
        
        *in their native quotation, pa, pb = -dy/dx at the y-intercept and x-intercept
        respectively; as the function y(x) is convex we must have pa >= pb; as this can
        be confusing in reverse quotation we correct by exchanging pa, pb if pb < pa
        """
        if pair.has_basetoken(tkn):
            pa = 1./pa
            pb = 1./pb
        
        if pa < pb:
            print("[from_prices] exchanging pa, pb")
            paa = pa
            pa = pb
            pb = paa
            
        B = sqrt(pb)
        S = sqrt(pa) - sqrt(pb)
        return cls(
            pair=pair, 
            tkn=tkn, 
            B=B, 
            S= S, 
            yint=yint, 
            y=y
        )
    
    @classmethod
    def from_order(cls, order):
        """
        alternative constructor, from an Order object

        :order:     the order object
        """
        return cls(
            pair=order.pair, 
            tkn=order.tkn, 
            B=float(order.B), 
            S= float(order.S), 
            yint=float(order.y_int), 
            y=float(order.y)
        )

    @property
    def p0(self):
        """
        the average or effective price of the range, p0 = sqrt(pa, pb)
        """
        return sqrt(self.pa*self.pb)
    
    @property
    def px(self):
        """alias for pb"""
        return self.pb 
    
    @property
    def py(self):
        """alias for pa"""
        return self.pa  
    
    @property
    def widthr(self):
        """the width ratio of the range, widthr = pmax/pmin"""
        return self.pmax/self.pmin
    
    @property
    def widthpc(self):
        """the percentage width of the range, widthpc = (pmax-pmin)/p0"""
        return (self.pmax-self.pmin)/self.p0

    @property
    def price_convention(self):
        """
        the price convention of the prices quoted
        
        :raw:   if False (default) return convention for all prices except 
                pa_raw and pb_raw which is returned for raw=True
        """
        if raw:
            return self.pair.price_convention(self.reverseq)
        return self.pair.price_convention()

    def descr(self, full=False):
        """provides a description of the order and curve"""
        s1 = f"Sell {self.tkn} buy {self.pair.other(self.tkn)}"
        s2 = f"from {self.pa:.4f} to {self.pb:.4f} {self.price_convention()}"
        s2 = f"from {self.pa:.4f} to {self.pb:.4f} {self.price_convention}"
        if full:
            s3 = f" ({self.y} {self.tkn} on curve, {self.y/self.yint*100:.0f}% of capacity)"
        else:
            s3 = ""
        return f"{s1} {s2}{s3}"

    @property
    def p_marg(self):
        """
        the current marginal price of the range
        """
        #dydx = ((self.B * self.yint + self.S * self.y) / self.yint)**2
        
        if self.yint == 0:
            if not self.y == 0:
                raise ValueError("If yint=0 you must also have y=0", yint, y)
            dydx = ((self.B + self.S))**2
        else:
            dydx = ((self.B + self.S * self.y/self.yint))**2
        result = dydx if self.pair.has_quotetoken(self.tkn) else 1/dydx
        return result
    
    @property
    def total_liquidity(self):
        """
        returns the total liquidity of the position, and the token in which it is quoted

        :returns:       (liquidity, token)
        """
        return self.y, self.tkn
    
    def liquidity_approx(self, price1, price2, tkn=None, asperc=False):
        """
        returns the approximate liquidity between start and end, in tkn

        :price1/2:      the start and end price of the range (in any order; quoted in price convention of pair)
        :tkn:           the token in which the liquidity is quoted (if None: base token)
        :asperc:        if True, return percentage total liquidity rather than tkn number; default is False
        :returns:       the liquidity in [price1, price2], quoted in tkn (or percent)
        """

        # ensure that price1 <= price2
        if price1 > price2:
            pp = price2
            price2 = price1
            price1 = pp

        if tkn is None:
            tkn = self.pair.basetoken
            #print(f"[liquidity_approx] token set to {tkn}")
            
        # # now we need to go through all the different arrangements of the range
        # # we first compute the percentage coverage, which simply is the percentage
            
        # # price1>pmax -> completely above the range
        # if price1 > self.pmax:
        #     perc = 0.
        
        # # price2<pmin -> completely below the range
        # elif price2 < self.pmin:
        #     perc = 0.

        # # price1 <= pmin
        # elif price1 <= self.pmin:

        #     # price1 <= min and price2 >= max -> 100%
        #     if price2 >= self.pmax:
        #         perc = 1.
        #     else:
        #         perc = (price2 - self.pmin)/(self.pmax - self.pmin)

        # # price2 >= pmax and price1>pmin
        # elif price2 >= self.pmax:

        #     perc = (self.pmax - price1)/(self.pmax - self.pmin)

        # # price2 < pmax and price1 > pmin
        # else:

        #     perc = (price2 - price1)/(self.pmax - self.pmin)

        # self.pmax == self.pmin -> 100% in range, otherwise out
        if self.pmax == self.pmin:
            if price1 == price2:
                perc2 = 1. if price1 == self.pmax else 0
            else:
                perc2 = 1. if price1 <= self.pmax and price2 > self.pmax else 0
            price = self.pmax
            #print(f"perc2={perc2}, price={price}", price1, price2, self.pmin, price1==self.pmax, price2==self.pmax)
        
        # price1 > self.pmax or price2 < self.pmin -> completely above or below the range
        elif price1 >= self.pmax or price2 <= self.pmin:
            perc2 = 0
        
        # alternatively: restrict p1,p2 to the ranges and calculate the coverage
        else:
            if price1 < self.pmin: price1 = self.pmin
            if price2 > self.pmax: price2 = self.pmax
            perc2 = (price2 - price1)/(self.pmax - self.pmin)
            price = sqrt(price1*price2)
        
        if asperc:
            return perc2
        
        if perc2 == 0:
            return 0

        # liquidity
        liq0 = self.y
        liq0_tkn = self.tkn

        # convert price into the correct quantity
        liq = self.pair.convert(amtfrom=liq0, tknfrom=liq0_tkn, tknto=tkn, price=price)
        
        return liq*perc2
        


"""
Exact-method router class using the x0, y0, n variables.

(c) Copyright Bprotocol foundation 2022.
Licensed under MIT
"""
from .exact_base_router import *


@dataclass
class ExactRouterX0Y0N(ExactBase):
    """
    Main class to handle the exact routing mothod. This method determines the ideal distribution of the input
    amount to arrive at **the** theoretical best price that can be obtained for a given set of orders.

    There are two versions of the exact method, which are equivilent in terms of the output, but differ in terms of
    variables used throughout. This version uses the x0, y0, n variables - which are not used in the smart-contract
    implementation or in the other simulator class ExactRouterSBY.
    """

    def amt_by_src(
        self, subject: int, dx: DecFloatInt, position_subset: List[int]
    ) -> DecFloatInt:
        """
        Calculates the liquidity amount change [delta-y] for y, via contributing delta-x (dx)
        """
        sub = self.orders[subject]
        n_sub, x0_sub, y_sub, y0_sub, x_sub, k_sub, a = (
            sub.n,
            sub.x0,
            sub.y,
            sub.y0,
            sub.x,
            sub.k,
            sub.a,
        )
        collections = self.handle_subfunctions_src(subject, position_subset, dx)
        b = self.b_src(x_sub, n_sub, x0_sub, k_sub, collections)
        e = collections["e_src"]
        f = collections["f_src"]

        self.logger.debug(
            "amt_by_src: a=%s, b=%s, e=%s, f=%s",
            self.frmt(a),
            self.frmt(b),
            self.frmt(e),
            self.frmt(f),
        )
        return (
            a**2 * b / (x0_sub * (a * b * n_sub + k_sub * x0_sub * (e * n_sub + f)))
        )

    def amt_by_target(
        self, subject: int, dx: DecFloatInt, position_subset: List[int]
    ) -> DecFloatInt:
        """
        Calculates the liquidity amount change [delta-y] for y, via contributing delta-x (dx)
        """
        sub = self.orders[subject]
        n, x0, y, y0, x, k, a_sub = (
            sub.n,
            sub.x0,
            sub.y,
            sub.y0,
            sub.x,
            sub.k,
            sub.a,
        )
        collections = self.handle_subfunctions_target(subject, position_subset)
        a = collections["a_target"]
        b = collections["b_target"]
        c = collections["c_target"]
        d = -(n * y + y0 * (1 - n))
        e = collections["e_target"]
        self.logger.debug(
            "amt_by_target: a=%s, b=%s, c=%s, d=%s, e=%s",
            self.frmt(a),
            self.frmt(b),
            self.frmt(c),
            self.frmt(d),
            self.frmt(e),
        )
        return (x0 * y0 * (a * (b + dx) + c) + d * e) / (n * e + x0 * y0 * a)

    def handle_subfunctions_src(
        self, subject: int, position_subset: List[int], amt: DecFloatInt
    ) -> dict:
        """
        Subfunction calculations for "c", "e", "f", "g" variables in the dy_by_src equation.
        """
        subject_free_indexes = [i for i in position_subset if i != subject]
        n, e_src, c_src, g_src = [], [], [], [-amt]
        for i in subject_free_indexes:
            n += [self.orders[i].n]
            g_src += [self.orders[i].x0]
            g_src += [-self.orders[i].x]
            n_collection, rs = [], {}
            sub_list = [j for j in subject_free_indexes if j != i]
            for j in sub_list:
                n_collection += [self.orders[j].n]
            e_src += [
                self.prod(n_collection)
                * Decimal.sqrt(self.orders[subject].k * self.orders[i].k)
            ]
            c_src += [self.prod(n_collection) * self.orders[i].x0]
        return {
            "c_src": sum(c_src),
            "e_src": sum(e_src),
            "f_src": self.orders[subject].k * self.prod(n),
            "g_src": sum(g_src),
        }

    def handle_subfunctions_target(
        self, subject: int, position_subset: List[int]
    ) -> dict:
        """
        Subfunction calculations for "a", "b", "c", "e" variables in the dy_by_target equation.
        """
        subject_free_indexes = [i for i in position_subset if i != subject]
        n_observers = [self.orders[i].n for i in subject_free_indexes]
        y_observers = [self.orders[i].y for i in subject_free_indexes]
        y0_observers = [self.orders[i].y0 for i in subject_free_indexes]
        rotation_sets = {}
        for j in subject_free_indexes:
            n_elements = [self.orders[k].n for k in subject_free_indexes if k != j]
            rotation_sets[j] = {
                "n_elements": n_elements,
                "x0_element": self.orders[j].x0,
                "y0_element": self.orders[j].y0,
            }
        x0, y0, y = (
            self.orders[subject].x0,
            self.orders[subject].y0,
            self.orders[subject].y,
        )
        return {
            "a_target": self.prod(n_observers),
            "b_target": sum(y_observers) - sum(y0_observers),
            "c_target": sum(
                [
                    self.prod(rotation_sets[i]["n_elements"])
                    * rotation_sets[i]["y0_element"]
                    for i in rotation_sets
                ]
            ),
            "e_target": sum(
                [
                    self.prod(rotation_sets[i]["n_elements"])
                    * Decimal.sqrt(
                        rotation_sets[i]["x0_element"]
                        * rotation_sets[i]["y0_element"]
                        * x0
                        * y0
                    )
                    for i in rotation_sets
                ]
            ),
        }

    def b_src(
        self,
        x: DecFloatInt,
        n: DecFloatInt,
        x0: DecFloatInt,
        k: DecFloatInt,
        collections: dict,
    ) -> DecFloatInt:
        """
        Subfunction calculation for "b" variable in the dy_by_src equation.
        """
        e = collections["e_src"]
        f = collections["f_src"]
        g = collections["g_src"]
        c = collections["c_src"]
        d = -n * x0 + n * x + x0
        self.logger.debug(
            "b_src: x=%s, n=%s, x0=%s, k=%s, e=%s, f=%s, g=%s, c=%s, d=%s",
            self.frmt(x),
            self.frmt(n),
            self.frmt(x0),
            self.frmt(k),
            self.frmt(e),
            self.frmt(f),
            self.frmt(g),
            self.frmt(c),
            self.frmt(d),
        )
        return -c * k + d * e + f * g
